/*
It looks like asyncLocalStorage only works when they all share an instance
of the AsyncLocalStorage class. This provides a mechanism to share a single
instance of asyncLocalStorage across files.
E.g.:

    //File1.js
    const asyncLocalStorage = new AsyncLocalStorage();
    
    const store = asyncLocalStorage.getStore();
    store.set('foo', 'bar');
    
    //File2.js
    const file1 = require('File1');
    
    const asyncLocalStorage = new AsyncLocalStorage();
    const store = asyncLocalStorage.getStore();
    
    store.get('foo') // does not work
 */
const { AsyncLocalStorage } = require("async_hooks");
const { isNil, isMap } = require("lodash");

const ROOT_KEY = "_AGNES_NAMESPACES_";
const DEFAULT_NAMESPACE = "_AGNES_DEFAULT_NAMESPACE_";
process[ROOT_KEY] = process[ROOT_KEY] || {}; // Copying from cls-hooked to handle the case where multiple instances of this file COULD be loaded

const getNamespace = (namespace = DEFAULT_NAMESPACE) => {
  let asyncLocalStorage = process[ROOT_KEY][namespace];
  if (isNil(asyncLocalStorage)) {
    asyncLocalStorage = new AsyncLocalStorage();
    process[ROOT_KEY][namespace] = asyncLocalStorage;
  }

  return asyncLocalStorage;
};

const destroyNamespace = (namespace = DEFAULT_NAMESPACE) => {
  const instance = getNamespace(namespace);
  instance.disable();
  process[ROOT_KEY][namespace] = null;
};

const reset = () => {
  if (process[ROOT_KEY]) {
    Object.keys(process[ROOT_KEY]).forEach(function (namespace) {
      destroyNamespace(namespace);
    });
  }
  process[ROOT_KEY] = Object.create(null);
};

const runNamespace = (fn, namespace = DEFAULT_NAMESPACE) => {
  const currentNamespace = getNamespace(namespace);
  return currentNamespace.run(currentNamespace.getStore() || new Map(), fn);
};
const run = (fn) => runNamespace(fn, DEFAULT_NAMESPACE);

const getStoreNamespace = (namespace = DEFAULT_NAMESPACE) =>
  getNamespace(namespace).getStore();
const getStore = () => getStoreNamespace(DEFAULT_NAMESPACE);

const setValueNamespace = (key, value, namespace = DEFAULT_NAMESPACE) =>
  getStoreNamespace(namespace).set(key, value);
const setValue = (key, value) =>
  setValueNamespace(key, value, DEFAULT_NAMESPACE);

const getValueNamespace = (key, namespace = DEFAULT_NAMESPACE) => {
  const storeNamespace = getStoreNamespace(namespace);
  if (isMap(storeNamespace)) {
    return storeNamespace.get(key);
  }
};
const getValue = (key) => getValueNamespace(key, DEFAULT_NAMESPACE);

module.exports = {
  getNamespace,
  destroyNamespace,
  reset,
  runNamespace,
  run,
  getStoreNamespace,
  getStore,
  setValueNamespace,
  setValue,
  getValueNamespace,
  getValue,
};
