const winston = require("winston");
const { format } = winston;

const path = require("path");
const getTraceId = require("../httpUtils/getTraceId");
const getLambdaContext = require("../httpUtils/getLambdaContext");
const { isNil, negate: not, isEmpty, isObject } = require("lodash");
const stackTrace = require("stack-trace");

const npmLogLevelKeys = require("../constants/npmLogLevelsKeys");
const { INFO, DEBUG } = npmLogLevelKeys;
const logKeyArray = Object.values(npmLogLevelKeys);

/**
 * There is an issue with Winston creating many EventEmitters. Winston v2 set maxListeners to 0 (infinite), but in v3
 * we must do it manually.
 *
 * Set to an arbitrary high number. The default is 10.
 *
 * @see https://github.com/winstonjs/winston/issues/1334
 * @see https://stackoverflow.com/questions/50709059/maxlistenersexceededwarning-possible-eventemitter-memory-leak-detected-11-mess
 */
require("events").EventEmitter.defaultMaxListeners = 100;

const errorHandler = new winston.transports.Console({
  format: format.combine(
    format.label({ label: "unhandled exceptions" }),
    format.timestamp(),
    format.simple()
  ),
});

/**
 * Include traceId and awsRequestId in the log output if they are present in the lambda context
 */
const traceIdFilter = format((info) => {
  const traceId = getTraceId();
  const awsRequestId = getLambdaContext.getAWSRequestId();
  if (isNil(traceId) && isNil(awsRequestId)) {
    return info;
  }

  // Add in the trace id and aws requestId if present
  let newMsg = "";
  if (!isNil(traceId)) {
    newMsg = `traceId [${traceId}]`;
  }
  if (!isNil(awsRequestId)) {
    if (!isNil(traceId)) {
      newMsg = newMsg + " ";
    }
    newMsg = newMsg + `awsRequestId [${awsRequestId}]`;
  }

  info.message = `${newMsg}: ${info.message}`;
  return info;
});

/**
 * Override default printing behaviour of Error objects passed as arguments to the `.log()` functions.
 *
 * @see https://github.com/winstonjs/winston#creating-custom-formats
 */
const errorFormatter = format((info) => {
  // Extra metadata is contained in the "splat" Symbol
  const splat = info[Symbol.for("splat")] || [];
  const error = splat.find((obj) => obj instanceof Error);

  if (error) {
    const errorMessage = error.message;

    // De-concatenate err.message from the log message
    // Default behaviour is to concat err.message and info.message, we don't want this.
    const logMessage = info.message.replace(` ${errorMessage}`, "");

    // Do not include "timings" from Got in the resulting error message
    const { message, timings, ...ownEnumerableProps } = error;

    // Append the error JSON to the end of the log message
    info.message = `${logMessage} ${JSON.stringify({
      message,
      ...ownEnumerableProps,
    })}`;
  }
  return info;
});

/**
 * If any of the arguments to `.log()` are Objects, convert them to a JSON string so they can be printed on a single line.
 *
 * @see https://github.com/winstonjs/winston#creating-custom-formats
 */
const jsonFormatter = format((info) => {
  if (isObject(info.message)) {
    info.message = JSON.stringify(info.message);
  }

  // Extra metadata is contained in the "splat" Symbol
  const splat = info[Symbol.for("splat")] || [];
  const json = splat.find((obj) => isObject(obj) && !(obj instanceof Error)); // Errors are handled separately

  if (json) {
    info.message = `${info.message} ${JSON.stringify(json)}`;
  }

  return info;
});

/**
 * Attempting to protect logging against mismatched case or typos.
 */
const getDefaultLoggerLevel = () => {
  if (!process.env.DEFAULT_LOGGER_LEVEL) {
    return INFO;
  }
  const proposedLevelLowerCase = process.env.DEFAULT_LOGGER_LEVEL.toLowerCase();

  if (!logKeyArray.includes(proposedLevelLowerCase)) {
    return INFO;
  }

  return proposedLevelLowerCase;
};

const getDebugLogCategories = () => {
  if (!process.env.DEBUG_LOG) {
    return [];
  } else {
    const debugCategoryList = process.env.DEBUG_LOG;
    return debugCategoryList
      .split(",")
      .filter(not(isEmpty))
      .map((debugCategory) => {
        return debugCategory.trim().toLowerCase();
      });
  }
};

const DEFAULT_LOGGER_LEVEL = getDefaultLoggerLevel();
const DEBUG_CATEGORIES_ARRAY = getDebugLogCategories();

let winstonLoggerConfig = null;

const getLogLevelForCategory = (logCategory = "") => {
  let level =
    winstonLoggerConfig != null ? winstonLoggerConfig[logCategory] : null;
  if (!level) {
    if (DEBUG_CATEGORIES_ARRAY.includes(logCategory.toLowerCase())) {
      level = DEBUG;
    } else {
      level = DEFAULT_LOGGER_LEVEL;
    }
  }
  return level;
};

/**
 * Fetch a logger instance from a category string.
 *
 * @example getLogger("myModuleLogger")
 *
 * @param {String} logCategory
 * @returns {winston.Logger} - Logger instance
 */
const getLogger = (logCategory) => {
  if (!winston.loggers.loggers[logCategory]) {
    winston.loggers.add(logCategory, {
      level: getLogLevelForCategory(logCategory),
      transports: [new winston.transports.Console()],
      format: format.combine(
        format.timestamp(),
        jsonFormatter(),
        errorFormatter(),
        traceIdFilter(),
        format.printf(
          (info) =>
            `${info.timestamp} - ${info.level}: [${logCategory}] ${info.message}`
        )
      ),
      exitOnError: false,
      exceptionHandlers: [errorHandler],
    });
  }
  return winston.loggers.get(logCategory);
};

/**
 * Fetch a logger instance using a name derived from the filepath or calling module.
 *
 * If the filepath is omitted, the path of the calling module will be used.
 *
 * @example
 * getLoggerFromFilepath();
 *
 * @example
 * getLoggerFromFilepath(__filename);
 *
 * @example
 * getLoggerFromFilepath('/path/to/another/file.js');
 *
 * @param {String} [filepath] - Filepath to parse
 * @returns {winston.Logger} - Logger instance
 */
const getLoggerFromFilepath = (filepath) => {
  let filename = "default";

  if (isNil(filepath)) {
    try {
      // Fetch the call stack
      const trace = stackTrace.get();
      // Fetch the filename of the previous function calls (i.e. not this one)
      filename = trace[1].getFileName();
    } catch (err) {
      // We can't do proper logging in the code to fetch the logger :-/
      console.log("Problem fetching filename from stack");
    }
  } else {
    filename = filepath;
  }

  // Split the filename from the path
  filename = path.basename(filename);

  // Split the filename from the suffix
  const logCategory = filename.split(".")[0];

  return getLogger(logCategory);
};

const logger = getLogger("ag-common-utils");

const updateLogLevel = (config) => {
  logger.info(
    "executing updateLogLevel based on config: " + JSON.stringify(config)
  );
  winstonLoggerConfig = config;

  // winston.loggers.loggers is a Map
  winston.loggers.loggers.forEach((logger, key) => {
    const level = getLogLevelForCategory(key);
    if (logger.level !== level) {
      logger.info("Changing log level of " + key + " to " + level);
      logger.level = level;
    }
  });
};

module.exports = {
  getLogger,
  getLoggerFromFilepath,
  updateLogLevel,
};
